/*
randomized random-cipher encryption
rrce License 200708231502

Copyright (c) 2007 Gregory Alan Hildstrom

All rights reserved.

Permission is hereby granted, free of charge, to any person (the
"User") obtaining a copy of this software and associated documentation
files (the "Software"), to deal in the Software without restriction,
including without limitation the rights to use, copy, modify, merge,
publish, distribute, and/or sell copies of the Software, and to permit
persons to whom the Software is furnished to do so, subject to the
following conditions:

1.  The above copyright notices and this permission notice (which
includes the disclaimer below) shall be included in all copies or
substantial portions of the Software.

2.  The name of a copyright holder shall not be used to endorse or
promote products derived from the Software without specific prior
written permission.

THIS DISCLAIMER OF WARRANTY CONSTITUTES AN ESSENTIAL PART OF THIS
LICENSE.  NO USE OF THE SOFTWARE IS AUTHORIZED HEREUNDER EXCEPT UNDER
THIS DISCLAIMER.  THE SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS
"AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING
BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A
PARTICULAR PURPOSE AND NONINFRINGEMENT OF THIRD PARTY RIGHTS.  IN NO
EVENT SHALL THE COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, OR ANY SPECIAL
INDIRECT OR CONSEQUENTIAL DAMAGES, OR ANY DAMAGES WHATSOEVER RESULTING
FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT,
NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION
WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.  NO ASSURANCES ARE
PROVIDED BY THE COPYRIGHT HOLDERS THAT THE SOFTWARE DOES NOT INFRINGE
THE PATENT OR OTHER INTELLECTUAL PROPERTY RIGHTS OF ANY OTHER ENTITY.
EACH COPYRIGHT HOLDER DISCLAIMS ANY LIABILITY TO THE USER FOR CLAIMS
BROUGHT BY ANY OTHER ENTITY BASED ON INFRINGEMENT OF INTELLECTUAL
PROPERTY RIGHTS OR OTHERWISE.  AS A CONDITION TO EXERCISING THE RIGHTS
GRANTED HEREUNDER, EACH USER HEREBY ASSUMES SOLE RESPONSIBILITY TO SECURE
ANY OTHER INTELLECTUAL PROPERTY RIGHTS NEEDED, IF ANY.  THE SOFTWARE
IS NOT FAULT-TOLERANT AND IS NOT INTENDED FOR USE IN MISSION-CRITICAL
SYSTEMS, SUCH AS THOSE USED IN THE OPERATION OF NUCLEAR FACILITIES,
AIRCRAFT NAVIGATION OR COMMUNICATION SYSTEMS, AIR TRAFFIC CONTROL
SYSTEMS, DIRECT LIFE SUPPORT MACHINES, OR WEAPONS SYSTEMS, IN WHICH
THE FAILURE OF THE SOFTWARE OR SYSTEM COULD LEAD DIRECTLY TO DEATH,
PERSONAL INJURY, OR SEVERE PHYSICAL OR ENVIRONMENTAL DAMAGE ("HIGH
RISK ACTIVITIES").  THE COPYRIGHT HOLDERS SPECIFICALLY DISCLAIM ANY
EXPRESS OR IMPLIED WARRANTY OF FITNESS FOR HIGH RISK ACTIVITIES.

end rrce License
*/

#include <iostream>
#include "rrce.h"
using namespace std;

int main(int argc, char** argv){
	if(argc != 5){
		cerr << "Error: rrce requires 4 arguments\n";
		cerr << " rrce <encrypt or decrypt> inputfile outputfile keyfile\n";
		exit(1);
	}//end if

	if((string)argv[1] != (string)"encrypt" && (string)argv[1] != (string)"decrypt"){
		cerr << "Error: check arguments, must choose encrypt or decrypt\n";
		exit(1);
	}//end if

	ifstream input;
	ofstream output;

	//open input file
	input.open(argv[2],ios::binary);
	if(!input.good()){
		cerr << "Error: could not open input file " << argv[2] << endl;
		exit(3);
	}//end if

	//check for existing output file and optionally open it
	ifstream ot;
	ot.open(argv[3]);
	if(ot.good()){
		cerr << "Error: output file " << argv[3] << " exists, overwrite? (Y/n): ";
		char choice = 'n';
		cin >> choice;
		if(choice != 'Y'){
			exit(3);
		}//end if
	}//end if
	ot.close();
	output.open(argv[3],ios::binary|ios::trunc);

	//create rrce object and load key
	rrce r;
	r.loadkey(argv[4]);

	//deal with initial offset and input filesize
	offsettype initialoffset = 0;
	time_t currenttime;
	unsigned long long filesize = 0;
	input.seekg(0,ios::end);
	filesize = input.tellg();
	if(filesize == 0){
		cerr << "Error: input file is empty.\n";
		exit(2);
	}//end if
	input.seekg(0,ios::beg);
	input.read((char*)&initialoffset,sizeof(initialoffset));//need to read first offsetsize bytes if encrypting or decrypting
	if((string)argv[1] == (string)"encrypt"){
		//compute initial offset and go back to beginning of input file
		time(&currenttime);
		initialoffset = currenttime;
		//initialoffset = 0;
		r.setinitialoffset(initialoffset);
		input.seekg(0,ios::beg);

		//write initial offset to output file
		output.write((char*)&initialoffset,sizeof(initialoffset));
	}//end if encrypt
	else{
		r.setinitialoffset(initialoffset);
		filesize -= sizeof(initialoffset);
	}//end else

	//create buffer
	char* buffer = NULL;
	unsigned long long buffersize = 1000000;
	buffer = (char*)malloc(buffersize);
	if(buffer == NULL){
		cerr << "Error: malloc buffer failed.\n";
		exit(2);
	}//end if

	//perform encryption or decryption
	unsigned long long readsize = 0;
	while(!input.eof()){
		//read input
		input.read(buffer,buffersize);
		if(!input.eof()) readsize = buffersize;
		else readsize = filesize % buffersize;

		//encrypt or decrypt
		if((string)argv[1] == (string)"encrypt")
			r.encrypt((unsigned char*)buffer,readsize);
		else
			r.decrypt((unsigned char*)buffer,readsize);

		//write output
		output.write(buffer,readsize);
		cout << "\r completed " << (unsigned short)((double)input.tellg()/(double)filesize*100) << "% " << input.tellg() << " out of " << filesize << " B"; cout.flush();
	}//end while
	initialoffset = 0;
	input.close();
	output.close();
	cout << "\r completed " << "100% " << filesize << " out of " << filesize << " B"; cout.flush();
	cout << endl;

	//clean up
	for(unsigned long long c=0; c < buffersize; c++)
		buffer[c] = 0;
	free(buffer);

	return 0;
}//end main

