/*
randomized random-cipher encryption
rrce License 200708231502

Copyright (c) 2007 Gregory Alan Hildstrom

All rights reserved.

Permission is hereby granted, free of charge, to any person (the
"User") obtaining a copy of this software and associated documentation
files (the "Software"), to deal in the Software without restriction,
including without limitation the rights to use, copy, modify, merge,
publish, distribute, and/or sell copies of the Software, and to permit
persons to whom the Software is furnished to do so, subject to the
following conditions:

1.  The above copyright notices and this permission notice (which
includes the disclaimer below) shall be included in all copies or
substantial portions of the Software.

2.  The name of a copyright holder shall not be used to endorse or
promote products derived from the Software without specific prior
written permission.

THIS DISCLAIMER OF WARRANTY CONSTITUTES AN ESSENTIAL PART OF THIS
LICENSE.  NO USE OF THE SOFTWARE IS AUTHORIZED HEREUNDER EXCEPT UNDER
THIS DISCLAIMER.  THE SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS
"AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING
BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A
PARTICULAR PURPOSE AND NONINFRINGEMENT OF THIRD PARTY RIGHTS.  IN NO
EVENT SHALL THE COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, OR ANY SPECIAL
INDIRECT OR CONSEQUENTIAL DAMAGES, OR ANY DAMAGES WHATSOEVER RESULTING
FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT,
NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION
WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.  NO ASSURANCES ARE
PROVIDED BY THE COPYRIGHT HOLDERS THAT THE SOFTWARE DOES NOT INFRINGE
THE PATENT OR OTHER INTELLECTUAL PROPERTY RIGHTS OF ANY OTHER ENTITY.
EACH COPYRIGHT HOLDER DISCLAIMS ANY LIABILITY TO THE USER FOR CLAIMS
BROUGHT BY ANY OTHER ENTITY BASED ON INFRINGEMENT OF INTELLECTUAL
PROPERTY RIGHTS OR OTHERWISE.  AS A CONDITION TO EXERCISING THE RIGHTS
GRANTED HEREUNDER, EACH USER HEREBY ASSUMES SOLE RESPONSIBILITY TO SECURE
ANY OTHER INTELLECTUAL PROPERTY RIGHTS NEEDED, IF ANY.  THE SOFTWARE
IS NOT FAULT-TOLERANT AND IS NOT INTENDED FOR USE IN MISSION-CRITICAL
SYSTEMS, SUCH AS THOSE USED IN THE OPERATION OF NUCLEAR FACILITIES,
AIRCRAFT NAVIGATION OR COMMUNICATION SYSTEMS, AIR TRAFFIC CONTROL
SYSTEMS, DIRECT LIFE SUPPORT MACHINES, OR WEAPONS SYSTEMS, IN WHICH
THE FAILURE OF THE SOFTWARE OR SYSTEM COULD LEAD DIRECTLY TO DEATH,
PERSONAL INJURY, OR SEVERE PHYSICAL OR ENVIRONMENTAL DAMAGE ("HIGH
RISK ACTIVITIES").  THE COPYRIGHT HOLDERS SPECIFICALLY DISCLAIM ANY
EXPRESS OR IMPLIED WARRANTY OF FITNESS FOR HIGH RISK ACTIVITIES.

end rrce License



randomized random-cipher encryption
encrypts 1 byte at a time by xor with 1 byte key segment from key file
initial key segment offset is determined from nonce, input data, and key file data
order of key segments is determined by key file data and encrypted input data

offset type can be 16, 32, or 64 bit
16bit 65536B (524288bit) max key size
32bit 4GB max key size
64bit TB max key size

initial offset is determined by using firstbytes (the first offsetsize bytes of the input file) with a nonce. The firstbytes are xored with the current time_t (32-bit), which yields firstbytesXORtime. The initial offset is computed by key[firstbytesXORtime] XOR key[key[firstbytesXORtime]]

--byte layout--
initial random key offset, which size depends on offsettype
8bits data xor with 8bits of key @ key[initial offset]; nextoffset = key[nextOffsetSizeChunk] xor current encrypted data
8bits data xor with 8bits of key @ offset; nextoffset = key[nextOffsetSizeChunk] xor current encrypted data
8bits data xor with 8bits of key @ offset
--end byte layout--

Switched from manual integer key location computation to pointers, which is almost twice as fast.

The output filesize = inputfilesize+offset type size.

The key offsets can be 64bits, which allows for a huge key file, but it revolves
for smaller key files. This makes it very difficult to determine key size.

The entire key is also not guaranteed to be used on any one file.

---Known Plaintext Attack, which already constitutes a security breach---
In the case of a known plaintext attack, segments of the key can be easily
determined by xoring the ciphertext and known plaintext. However, the order and
extent of key segments is unknown, so the original key cannot be easily reconstructed.

In the case that a set of key segments is determined from a known plaintext
attack and another encrypted file exists with the same initial key offset,
additional data are kept safe by the computation of offsets from unknown
key data xored with ciphertext data. This ensures that neither the same
key segments nor the same order are used on different files starting with the same offset.

Known plaintext attack works iff (if and only if) the entire key was used, key size
is known, and all segments were compromised; otherwise permutations increase.
Assume a very small key size of 64 Bytes (512bits) for this example.
Key segment size is 1byte (8bits), but segment can start at any byte in key.
Number of key segments possible: 64.
All possible keys from different segment ordering: 64! ~= 1.27x10^89

Now imagine a key that is 100B (800bits).
Key segment size is still 1byte (8bits), but segment can start at any byte in key.
Now try to determine the correct key based on the order of the 100 segments
All possible keys from different segment ordering: 100! ~= 9.33x10^157

Now imagine a 1MB or 1GB key.

---Known Plaintext Attack, which already constitutes a security breach---


There may very well be a shortcut method to breaking this encryption, but I have not found it yet.


*/


#ifndef RRCE_H
#define RRCE_H

#include <string>
#include <fstream>
#include <time.h>
using namespace std;

//typedef unsigned long long offsettype;//64-bit OS and multi-terabyte max keys
typedef unsigned int offsettype;//32-bit OS and 4GB max keys
//typedef unsigned short offsettype;//16-bit OS and 65KB max keys

offsettype pow(offsettype number,offsettype power){
	offsettype output = 1;
	for(offsettype i=0; i < power; i++)
		output *= number;
	return output;
}

class rrce{
public:
	rrce();//default constructor
	~rrce();//default destructor
	void loadkey(char*);//loads secret encryption key, must be over 128bits or 16bytes
	offsettype computeinitialoffset(offsettype);//computes initial key offset from input plaintext data and key data (encryption)
	void setinitialoffset(offsettype);//sets initial key offset from input ciphertext data (decryption)
	void encrypt(unsigned char*,unsigned long long);//encrypts a buffer, assuming key loaded and initial offset computed
	void decrypt(unsigned char*,unsigned long long);//decrypts a buffer, assuming key loaded and initial offset read
private:
	unsigned char* key;
	offsettype keysize;
	unsigned char byte;
	unsigned char byte2;
	offsettype initialoffset;
	offsettype* offset1;
	offsettype* offset2;
	offsettype offset3;
	offsettype sizeofoffset;
	offsettype power;
	bool debug;
};//end class rrce

rrce::rrce(){
	debug = false;;
	sizeofoffset = sizeof(offsettype);
	key = NULL;
	keysize = 0;
	byte = 0;
	byte2 = 0;
	initialoffset = 0;
	offset1 = 0;
	offset2 = 0;
	offset3 = 0;
}//constructor

rrce::~rrce(){
	for(offsettype i=0; i < keysize+sizeofoffset; i++){
		key[i] = 0x00;
	}//end for
	free(key);
	keysize = 0;
	initialoffset = 0;
	offset1 = 0;
	offset2 = 0;
	offset3 = 0;
	byte = 0;
	byte2 = 0;
}//destructor

//load secret encryption key
void rrce::loadkey(char* keyfilename){
	ifstream keyfile;
	keyfile.open(keyfilename,ios::binary);
	if(!keyfile.good()){
		cerr << "Error: could not load key file " << keyfilename << endl;
		exit(1);
	}//end if
	keyfile.seekg(0,ios::end);
	keysize = keyfile.tellg();
	keyfile.seekg(0,ios::beg);
	key = NULL;
	key = (unsigned char*)malloc(keysize+sizeofoffset);
	if(key == NULL){
		cerr << "Error: key malloc failed for keysize " << keysize << endl;
		exit(2);
	}//end if
	keyfile.read((char*)key,keysize);
	keyfile.close();
	for(offsettype i=0; i < keysize; i++)
		if(key[i] == 0) key[i] = 0x27; // if some bits would not have been flipped, flip some anyway
	if(keysize < 16){
		cerr << "Error: keysize is too small!\n";
		exit(4);
	}//end if
	for(offsettype i=0; i < sizeofoffset; i++)//allow for easy revolving behavior
		key[keysize+i] = key[i];
	power = 1;
	offsettype tempval=0;
	while(tempval < keysize){
		power++;
		tempval = 1<<power;
	}
	if(tempval != keysize){
		cerr << "Error: keysize is not a power of 2\n";
		exit(4);
	}
	cout << "keysize is 2^" << power << endl;
	cout << "sizeof(offsettype) is " << sizeof(offsettype) << endl;
}//end loadkey

//set offset1 = key[initialoffset], which means that the initial offset is actually a pointer to a pointer used to encrypt the first byte
void rrce::setinitialoffset(offsettype initialoffset){
	offset2 = (offsettype*)&key[initialoffset%keysize];
	offset1 = (offsettype*)&key[(*offset2)%keysize];
}//end setinitialoffset

//encrypt
void rrce::encrypt(unsigned char* buffer, unsigned long long buffersize){
	offsettype d,r;
	for(unsigned long long i=0; i < buffersize; i++){
		//read input byte
		byte = buffer[i];

		//xor encrypt byte from key data at offset
		//print out key order
		byte ^= *(unsigned char*)offset1;

		//write output byte
		buffer[i] = byte;

		//compute next offset from key data and encrypted input byte
		offset2 = offset1 + 1;
		offset3 = (*offset2)^(offsettype)byte;
		////slow formula works for any key size
		//offset1 = (offsettype*)&key[offset3%keysize];
		////fast formula works for power of 2 key size
		d = offset3>>power;
		d <<= power;
		r = offset3-d;
		offset1 = (offsettype*)&key[r];
	}//end for
}//end encrypt

//decrypt
void rrce::decrypt(unsigned char* buffer, unsigned long long buffersize){
	offsettype d,r;
	for(unsigned long long i=0; i < buffersize; i++){
		//read input byte
		byte = buffer[i];

		//xor decrypt byte from key data at offset
		byte2 = byte;
		byte ^= *(unsigned char*)offset1;

		//write output byte
		buffer[i] = byte;

		//compute next offset from key data and encrypted input byte
		offset2 = offset1 + 1;
		offset3 = (*offset2)^(offsettype)byte2;
		////slow formula works for any key size
		//offset1 = (offsettype*)&key[offset3%keysize];
		////fast formula works for power of 2 key size
		d = offset3>>power;
		d <<= power;
		r = offset3-d;
		offset1 = (offsettype*)&key[r];
	
	}//end for
}//end decrypt


#endif

